# -*- encoding: utf-8 -*-
'''
Projects -> File: -> timedomainwav.py
Author: DYJ
Date: 2022/07/19 14:25:22
Desc: 时域信号类
      在时域内计算的指标放在这个类里面计算，对这个类调用频域内计算的指标会自动转化为对该类进行默认的加窗傅里叶变换后再计算
version: 1.0
To do: 与目前系统里的计算方式对比，看结果是否有差异，看实现是否正确
'''


import warnings
import numpy as np
import matplotlib.pyplot as plt
from scipy import signal
from PyEMD import EMD

from .rollingwav import RollingWav

from .basewav import BaseWav
from . import frequencydomainwav as fdwav


class TimeDomainWav(BaseWav):
    '''
    时域信号

    在时域内计算的指标放在这个类里面计算，对这个类调用频域内计算的指标会自动转化为对该类进行默认的加窗傅里叶变换后再计算
    '''
    def __init__(self, values, sample_frequency):
        """ 
        初始化TDWav, 时域信号初始化必须提供采样率

        Parameters
        ----------
        values : 1D list or array-like
        Wav的值序列
        sample_frequency : int
        采样率, 单位: Hz

        Returns
        -------
        None
        """
        self.values = np.asarray(values).squeeze()
        if len(self.values.shape) > 1:
            raise Exception('Wave data can not have more than 1-dimension!')
        self.length = len(self.values)
        self.sample_frequency = sample_frequency
        self.time_length = self.length / self.sample_frequency

        # alias
        self.frequency = self.sample_frequency

    # 当需要计算的指标或进行的变换在时域内无法计算时，会自动转向该时域信号以默认参数进行的加窗傅里叶变换后的频域信号进行计算
    # 如果需要以非默认参数进行的傅里叶变换，请使用reset_freq_domain函数重置频域；或者直接在频域进行计算。
    def __getattr__(self, name):
        if name in ['FC', 'MSF', 'VF', 'RMSF', 'RVF']:  # 这些指标的计算应该在功率谱上进行计算
            return getattr(self.psd_spectrum, name)
        # lazy-evaluation, 惰性求值，知道被访问到时再进行计算而不是在初始化时就计算
        elif name == "freq_domain": 
            freq_domain = self.fft()
            setattr(self, name, freq_domain)
            return freq_domain
        elif name == "psd_spectrum":
            psd_spectrum = self.PSD_Periodogram()
            setattr(self, name, psd_spectrum)
            return psd_spectrum
        else:
            return getattr(self.freq_domain, name)

    def reset_freq_domain(self, window='hann', convert2real=True):
        '''
        重置频域
        如果不想使用默认的傅里叶变换参数的结果进行后续频域指标计算, 可以通过这个函数重新计算
        
        Parameters
        ----------
        window: str or tuple
        指定傅里叶变换时使用的窗函数, 该参数将被用于调用signal.get_window
        convert2real: bool
        是否要将傅里叶变换的结果变为实数(取模)
        
        Return
        ------
        None
        '''
        self.freq_domain = self.fft(window, convert2real)
        return None

    # 绘图函数
    def plot(self, *args, **kwargs):
        plt.plot(np.arange(self.length) / self.sample_frequency, self.values, *args, **kwargs)
        plt.xlabel('Time(s)')
        return None

    # 在时域上进行的变换
    def fft(self, window='hann', convert2real=True):
        '''
        在时域内进行加窗傅里叶变换，需要除以数据长度

        Parameters
        ----------
        window : str
            加窗计算所使用的窗的名称, 支持scipy.signal.get_window内的所有窗
        convert2real : bool
            是否将傅里叶变换的结果从复数转化为实数(计算模)
        Returns
        -------
        fft_wav: FrequencyDomainWav
            傅里叶变换后的频域信号
        '''
        # 加窗抑制频谱泄露
        window = signal.get_window(window, self.length)
        windowed_value = self.values * window
        adjust_coefficient = round(1 / np.mean(window), 4)
        if convert2real:
            fft_value = np.abs(np.fft.fft(windowed_value)) / self.length * adjust_coefficient
        else:
            fft_value = np.fft.fft(windowed_value) / self.length * adjust_coefficient
        
        # 计算频率
        freq = np.arange(self.length) / self.length * self.sample_frequency
        fft_wav = fdwav.FrequencyDomainWav(fft_value, freq)[:self.length // 2]  # 只取傅里叶变换结果的前一半
        return fft_wav

    # 频谱分析/功率谱分析/功率谱密度
    # 1. 周期图法/直接法
    def PSD_Periodogram(self, window='hann'):
        """ 
        直接法求功率谱密度
        直接法又称周期图法(Periodogram)

        Parameters
        ----------
        window: str or tuple
        指定傅里叶变换时使用的窗函数, 该参数将被用于调用signal.get_window

        Returns
        -------
        psd_spectrum : FrequencyDomainWav
        """

        freq, spectrum = signal.periodogram(self.values, self.sample_frequency, window)
        psd_spectrum = fdwav.FrequencyDomainWav(spectrum, freq)
        return psd_spectrum

    # 2. Welch法/改进的直接法(加窗)
    def PSD_Welch(self, window='hann', nperseg=8192):
        '''
        用Welch法计算功率谱密度 PSD(Power Spectral Density)
        Welch法是改进的直接法, 先将信号分段加窗后计算功率谱, 随后对每段数据的功率谱进行平均

        Parameters
        ----------
        window: str or tuple
        指定傅里叶变换时使用的窗函数, 该参数将被用于调用signal.get_window
        npserseg: int
        每个window中含有的数据点的个数/window的长度

        Returns
        -------
        psd_spectrum : FrequencyDomainWav
        '''

        freq, spectrum = signal.welch(self.values, self.sample_frequency, window, nperseg)
        psd_spectrum = fdwav.FrequencyDomainWav(spectrum, freq)
        return psd_spectrum

    # 3. 间接法, 自相关函数法
    def PSD_autocorrelation(self, window='hann'):
        '''
        用自相关法计算功率谱密度 PSD(Power Spectral Density)
        自相关法是先计算序列的自相关系数, 然后对自相关系数进行傅里叶变换来得到功率谱

        Parameters
        ----------
        window: str or tuple
        指定傅里叶变换时使用的窗函数, 该参数将被用于调用signal.get_window

        Returns
        -------
        psd_spectrum : FrequencyDomainWav
        '''
        # 计算自相关的前提是函数f(t)的均值m(t)为零，因此需要先将数据demean
        demean_dat = self.values - self.Mean
        xcorr = np.correlate(demean_dat, demean_dat, mode="same")
        # 加窗抑制频谱泄露
        windowed_xcorr = xcorr * signal.get_window(window, len(xcorr))
        spectrum = np.abs(np.fft.fft(windowed_xcorr)) / self.length
        
        # 计算频率
        freq = np.arange(self.length) * self.sample_frequency / (self.length)
        psd_spectrum = fdwav.FrequencyDomainWav(spectrum, freq)
        return psd_spectrum

    # 熵
    # 1. 近似熵(https://zhuanlan.zhihu.com/p/494761890/)
    def calc_approximate_entropy(self, pattern_dimension=2, similarity_tolerance=0.20):
            """
            计算数据的近似熵

            Parameters  :
            ----------
            pattern_dimension: int
            模式维数
            similarity_tolerance: float
            相似容限阈值

            Returns  :
            -------
            approximate_entropy: float
            近似熵
            """
            if self.length > 5000 or self.length < 200:
                warnings.warn(f'The length of data used for approximate entropy should be 200~5000, not {self.length}')
            approximate_entropy = self._calc_phi(pattern_dimension, similarity_tolerance) - self._calc_phi(pattern_dimension + 1, similarity_tolerance)
            return  approximate_entropy

    def _calc_phi(self, pattern_dimension=2, similarity_tolerance=0.20):
        """
        给定数据，计算近似熵的中间量phi
        phi可以理解为在特定模式维数下，数据模式的多样性的衡量（越负越多样）
        Parameters  :
        ----------
        pattern_dimension: int
        模式维数
        similarity_tolerance: float
        相似容限阈值

        Returns  :
        -------
        phi: float
        """
        # 构造信号序列
        sig_list = np.asarray([self.values[i:i + pattern_dimension] for i in range(self.length - pattern_dimension + 1)])
        # 依次比较序列是否相似（即两个序列的对应端点的值相差均小于容差），统计相似数量，计算相似比例
        similar_ratio = []
        for seq in sig_list:
            diff_smaller_than_threshold = (sig_list - seq) <= (similarity_tolerance * self.SD)
            is_similar = diff_smaller_than_threshold.sum(axis=1) == pattern_dimension
            similar_ratio.append(is_similar.mean())
        # 计算phi
        similar_ratio = np.asarray(similar_ratio)
        phi = np.sum(np.log(similar_ratio)) / (self.length - pattern_dimension + 1)
        return phi

    # 经验模态分解
    def EMD(self):
        """
        经验模态分解(Empirical Mode Decomposition)

        Parameters  :
        ----------
        
        Returns  :
        -------
        (imfs, res): (ndarray, 1d-array)
        内涵模态分量(Intrinsic Mode Functions, IMF)的矩阵, 每一行代表一个分量
        
        To do: 把输出调整成wavesis的风格(输出成值、wav类或wav类的列表)
        """
        # 直接调用pyEMD库的算法
        pyemd = EMD()
        pyemd.emd(self.values)
        imfs, res = pyemd.get_imfs_and_residue()
        return imfs, res

    # 时频分析
    def time_frequency_analysis(self, NFFT=256, Fs=8000, mode='psd'):
        spectrum, freqs, times = plt.specgram(self.values, NFFT, Fs, noverlap=NFFT // 2, mode=mode)
        return freqs, times, spectrum

# To do: 测试代码
if __name__ == "__main__":
    pass


# -*- encoding: utf-8 -*-
'''
Projects -> File: -> rollingwav.py
Author: DYJ
Date: 2022/07/19 14:27:01
Desc: 滑动窗类, 实现Wav类的窗划分计算和聚合
version: 1.0
'''

from inspect import ismethod
from pdb import set_trace
import numpy as np

# To do: 对于特别耗时的计算，改写成view_as_windows的版本，提高运算效率
# from skimage.util import view_as_windows


class RollingWav(object):
    '''
    滑动窗类, 实现Wav类的窗滑动计算和聚合
    '''
    def __init__(self, Wav, window_width, step=1) -> None:
        self.Wav = Wav
        self.window_width = window_width
        self.step = step
        if self.step == 1:
            self.total_n_step = int(self.Wav.length - self.window_width + 1)
        else :
            self.total_n_step = int(np.floor((self.Wav.length - self.window_width) / self.step) + 1)

    def get_window_wav(self):
        for i_step in range(self.total_n_step):
            if self.Wav[i_step * self.step : i_step * self.step + self.window_width] is not None:
                yield self.Wav[i_step * self.step : i_step * self.step + self.window_width]
            else:
                raise StopIteration

    def __getattr__(self, name):
        window_wav = self.get_window_wav()
        res = [getattr(i_wav, name) for i_wav in window_wav]
        if not ismethod(getattr(self.Wav, name)): # 对于函数属性，直接存储为结果
            try:
                res_wav = self.Wav.__class__(res, self.Wav.frequency / self.step)
            except Exception:
                res_wav = res
        else: # 对于函数方法，先保存为RollingWav的一个属性，再随后对RollingWav的调用中，再将方法参数输入
            self.funcs = res
            res_wav = self
        return res_wav
    
    # 对于函数方法的rolling调用，参数在这里输入
    # 由于函数方法的输出多种多样，因此这里的输出不强制转换为Wav，而是直接输出列表
    def __call__(self, *args, **kwargs):
        res = [func(*args, **kwargs) for func in self.funcs] 
        return res

    def apply(self, func, *args, **kwargs):
        window_wav = self.get_window_wav()
        res = [func(wav.values, *args, **kwargs) for wav in window_wav]
        return res

class RollingWavBundle(object):
    '''
    滑动窗类, 实现WavBundle类的窗滑动计算和聚合
    WavBundle的基础运算方式，是对Bundle里面的每一个Wav进行相同的运算
        对于Wav的属性调用，直接用.调用
        对于Wav的函数调用，统一使用apply函数
    WavBundle的高级运算是需要多个波结合起来运算，比如三相电的park矢量变换和对称分量法
        这类计算统一直接使用WavBundle的函数调用，即使用.调用
    '''
    def __init__(self, WavBundle, window_width, step=1) -> None:
        self.WavBundle = WavBundle
        self.window_width = window_width
        self.step = step

    def get_window_wavbundle(self):
        if isinstance(self.WavBundle.shape[1], (list, tuple)):
            wav_length = min(self.WavBundle.shape[1])
        else:
            wav_length = self.WavBundle.shape[1]
        if self.step == 1:
            self.total_n_step = int(wav_length - self.window_width + 1) # 如果Bundle里面的Wav长度各不相同，去最短的长度进行切分
        else :
            self.total_n_step = int(np.floor((wav_length - self.window_width) / self.step) + 1)
        for i_step in range(self.total_n_step):
            if self.WavBundle[i_step * self.step : i_step * self.step + self.window_width] is not None:
                yield self.WavBundle[i_step * self.step : i_step * self.step + self.window_width]
            else:
                raise StopIteration

    def __getattr__(self, name):
        # To do： 思考如何把这一部分跟rollingwav的统一起来，应该是可以统一的
        if ismethod(getattr(self.WavBundle, name)): # 对于函数方法，将先保存在self.funcs里，在随后的__call__中传入参数并调用
            window_wavbundle = self.get_window_wavbundle()
            self.funcs = [getattr(i_wavbundle, name) for i_wavbundle in window_wavbundle]
            return self
        else: # 对于函数属性，直接调用
            res = [getattr(i_wav.rolling(self.window_width, self.step), name) for i_wav in self.WavBundle.wavs]
            return self.WavBundle.__class__(**dict(zip(self.WavBundle.wavnames, res)))   

    # 对于函数方法的rolling调用，参数在这里输入
    # 由于函数方法的输出多种多样，因此这里的输出不强制转换为Wav，而是直接输出列表
    def __call__(self, *args, **kwargs):
        res = [func(*args, **kwargs) for func in self.funcs] 
        return res

    def apply(self, func, *args, **kwargs):
        window_wavbundle = self.get_window_wavbundle()
        res = [i_wavbundle.apply(func, *args, **kwargs) for i_wavbundle in window_wavbundle]
        # return RollingWavBundle(**dict(zip(self.WavBundle.wavnames, res)))
        return res


# -*- encoding: utf-8 -*-
'''
Projects -> File: -> threephasewav.py
Author: DYJ
Date: 2022/07/27 16:06:49
Desc: 三相电类
version: 1.0
'''


from inspect import ismethod
import numpy as np
import matplotlib.pyplot as plt

from .basewav import BaseWav
from .rollingwav import RollingWav, RollingWavBundle
from . import frequencydomainwav as fdwav
from . import timedomainwav as tdwav

class WavBundle(object):
    '''
    波束类
    把多个波“捆”成束进行统一处理，即对多个波进行统一的操作
    WavBundle的基础运算方式，是对Bundle里面的每一个Wav进行相同的运算
        对于Wav的属性调用，直接用.调用
        对于Wav的函数调用，统一使用apply函数
    WavBundle的高级运算是需要多个波结合起来运算，比如三相电的park矢量变换和对称分量法
        这类计算统一直接使用WavBundle的函数调用，即使用.调用
    '''
    def __init__(self, **kwargs) -> None:
        '''
        WavBundle的初始化按照WavName=Wav的方式任意长度输入
        WavName即为波的名称，如'A相电流'
        Wav应该是【BaseWav, fdwav.FrequencyDomainWav, tdwav.TimeDomainWav, RollingWav】中的任意一种
        '''
        # check data
        for k,v in kwargs.items():
            assert isinstance(v, (BaseWav, fdwav.FrequencyDomainWav, tdwav.TimeDomainWav, RollingWav)), k + 'is not a Wav, please check it!'
        self.wavnames = [key for key in kwargs.keys()]
        self.wavs = [value for value in kwargs.values()]
        self.length = len(self.wavs)
        self.width = [len(i_wav) for i_wav in self.wavs]
        if len(set(self.width)) == 1:
            self.width = self.width[0]
    
    @staticmethod
    def init_unnamed_wavs_from_list(li):
        return WavBundle(**{'wav_' + str(i): i_wav for i, i_wav in enumerate(li)})

    @property
    def shape(self):
        return self.length, self.width

    def __getattr__(self, name):
        if not ismethod(getattr(self.wavs[0], name)): # 对于函数属性，直接存储为结果
            res = [getattr(i_wav, name) for i_wav in self.wavs]
        elif name in self.__dict__: # 对于函数方法，只能调用WavBundle自身的实现，否则引发错误
            res = getattr(self, name)
        else:
            raise ValueError(self.__class__.__name__ + 'dose not have function: ' + name + ', please use apply instead.')
        return res

    # 直接打印时，显示wavs
    def __repr__(self) -> str:
        res = [str(i_wav) for i_wav in self.wavs]
        return '\n'.join(res)

    # 切片功能
    def __getitem__(self, item): 
        cls = type(self)
        if isinstance(item, slice):
            res = [i_wav[item] for i_wav in self.wavs]
            return self.__class__(**dict(zip(self.wavnames, res)))
        elif isinstance(item, int):
            res = [i_wav[item] for i_wav in self.wavs]
            return res
        else:
            raise ValueError(str(item) + 'must be slice or int!')

    def apply(self, func, *args, **kwargs):
        res = [func(wav, *args, **kwargs) for wav in self.wavs]
        try:
            res = np.asarray(res) if isinstance(res[0], (tuple,)) else self.__class__(**dict(zip(self.wavnames, res)))
        finally:
            return res

    # 实现滑动窗计算，每个滑动窗都是一个WavBundle，可以在每个滑动窗进行所有指标的计算和转换
    def rolling(self, window_width, step=1):
        '''
        Parameters
        ----------
        window_width : int 
        滑动窗的宽度
        step : int
        滑动窗的步长

        Returns
        -------
        rolling_wavbundle: RollingWavBundle
        A generator of Wav
        '''
        rolling_wavbundle = RollingWavBundle(self, window_width, step)
        return rolling_wavbundle

class threephasewav(WavBundle):
    '''
    三相电类
    '''
    def __init__(self, ia=None, ib=None, ic=None, ua=None, ub=None, uc=None, 
                    sample_frequency=8000, wiring_structure=None) -> None:
        """
        初始化三相电类
        Parameters  :
        ----------
        ia, ib, ic: list or 1D-array
        相电流
        ua, uv, uc: list or 1D-array
        相电压
        wiring_structure: str
        接线方式："Y"(星形接线)或者"N"(三角接线)

        Returns  :
        -------
        None
        """
        self.data = {'ia':ia, 'ib':ib, 'ic':ic,
                     'ua':ua, 'ub':ub, 'uc':uc}
        self.sample_frequency = sample_frequency
        self.wiring_structure = wiring_structure
        # self.data_check = self._data_check()
        # if self.data_check == '数据不完整':
        #     raise Exception('电压电流数据均不完整，无法进行后续计算!')
        for k,v in self.data.items():
            if not isinstance(v, (BaseWav, RollingWav)) and not v is None:
                self.data[k] = tdwav.TimeDomainWav(v, self.sample_frequency)
        WavBundle.__init__(self, **{k:v for k,v in self.data.items() if v is not None})

    @staticmethod
    def init_wavs_from_rawdata(data, colname=['A相电流', 'B相电流', 'C相电流'], sample_frequency=8000, with_voltage=False, voltage_colname=['A相电压', 'B相电压', 'C相电压']):
        if with_voltage:
            currentA = tdwav.TimeDomainWav(data[colname[0]], sample_frequency=sample_frequency)
            currentB = tdwav.TimeDomainWav(data[colname[1]], sample_frequency=sample_frequency)
            currentC = tdwav.TimeDomainWav(data[colname[2]], sample_frequency=sample_frequency)
            voltageA = tdwav.TimeDomainWav(data[voltage_colname[0]], sample_frequency=sample_frequency)
            voltageB = tdwav.TimeDomainWav(data[voltage_colname[1]], sample_frequency=sample_frequency)
            voltageC = tdwav.TimeDomainWav(data[voltage_colname[2]], sample_frequency=sample_frequency)
            return threephasewav(ia=currentA, ib=currentB, ic=currentC, ua=voltageA, ub=voltageB, uc=voltageC)
        else:
            currentA = tdwav.TimeDomainWav(data[colname[0]], sample_frequency=sample_frequency)
            currentB = tdwav.TimeDomainWav(data[colname[1]], sample_frequency=sample_frequency)
            currentC = tdwav.TimeDomainWav(data[colname[2]], sample_frequency=sample_frequency)
            return threephasewav(ia=currentA, ib=currentB, ic=currentC)

    def _data_check(self):
        # 检查电流数据是否完整
        if self.data['ia'] and self.data['ib'] and self.data['ic']:
            current_ok = True
        else:
            current_ok = False
        # 检查电压数据是否完整
        if self.data['ua'] and self.data['ub'] and self.data['uc']:
            voltage_ok = True
        else:
            voltage_ok = False
        if current_ok and voltage_ok:
            res = '电流电压数据完整'
        elif current_ok:
            res = '电流数据完整'
        elif voltage_ok:
            res = '电压数据完整'
        else:
            res = '数据不完整'
        return res

    def park_transform(self, magnitude=False):
        """park矢量变换，将原始电信号转换为i_d，i_q

        Parameters
        ----------
        magnitude : bool
            是否取模

        Returns
        -------
        wavs : WavBundle(i_d, i_q)
            park矢量转换后的数据
        or wav: TimeDomainWav
            park矢量模

        """
        ia, ib, ic = np.asarray(self.data['ia']), np.array(self.data['ib']), np.array(self.data['ic'])
        i_d = np.sqrt(2 / 3) * ia - np.sqrt(1 / 6) * ib - np.sqrt(1 / 6) * ic
        i_q = np.sqrt(1 / 2) * ib - np.sqrt(1 / 2) * ic
        if magnitude:
            return tdwav.TimeDomainWav(np.sqrt(np.power(i_d, 2) + np.power(i_q, 2)), self.sample_frequency)
        else:
            return WavBundle(i_d=tdwav.TimeDomainWav(i_d, self.sample_frequency), i_q=tdwav.TimeDomainWav(i_q, self.sample_frequency))

    def plot(self):
        self.data['ia'].plot('r-', label='A相电流')
        self.data['ib'].plot('g-', label='B相电流')
        self.data['ic'].plot('b-', label='C相电流')
        plt.legend()
        return None
    # To do: 对称分量法
    # To do: 相位不平衡算法；幅值不平衡算法

if __name__ == '__main__':
    t = np.arange(0, 10, 0.01)
    phaseA = np.cos(2 * np.pi * t) * np.cos(100 * np.pi * t)
    phaseB = np.cos(2 * np.pi * t + np.pi * 2 / 3) * np.cos(100 * np.pi * t + np.pi * 2 / 3)
    phaseA_wav = tdwav.TimeDomainWav(phaseA)
    phaseB_wav = tdwav.TimeDomainWav(phaseB)
    wav_bundle = WavBundle(A=phaseA_wav, B=phaseB_wav)

    wav_bundle.fft()



# -*- encoding: utf-8 -*-
'''
Projects -> File: wavesis -> basewav.py
Author: DYJ
Date: 2022/07/19 12:14:09
Desc: 基础波类/基础振动信号类 
      该类实现所有不依赖时间信息或频率信息的指标计算和变换      
      注: 此处的振动是指自然界中广泛存在的物质系统的一种普遍运动形式, 通常指一个物理量在某一平衡值附近随时间而往复变化的过程 ———— 徐平《振动信号处理与数据分析》
version: 1.0
'''


import numpy as np
import matplotlib.pyplot as plt
from scipy import signal
from .rollingwav import RollingWav
from functools import cached_property

# Base Class
class BaseWav(object):
    '''
    基础波类，基础振动信号类
    '''
    def __init__(self, values, frequency=None):
        """ 
        初始化Wav

        Parameters
        ----------
        values : 1D list or array-like
        值序列/信号序列
        frequency: int or 1D list
        频率, 可选参数

        Returns
        -------
        None
        """
        self.values = np.asarray(values).squeeze()
        # 检查初始化数据
        if len(self.values.shape) > 1:
            raise Exception('Wave data can not have more than 1-dimension!') # todo： 考虑要不要把这个改成装饰器
        self.length = len(self.values)
        self.frequency = frequency

    # 直接打印时，显示振动信号的值
    def __repr__(self) -> str:
        return str(self.values)

    # 实现切片功能，对Wav类进行切片仍然得到Wav类
    def __getitem__(self, item): 
        cls = type(self) 
        if isinstance(item, slice):
            if self.frequency is None or isinstance(self.frequency, (int, float)):
                return self.__class__(self.values[item], self.frequency)
            elif len(self.frequency) > 1:
                return self.__class__(self.values[item], self.frequency[item])
            else :
                Warning('The type of self.frequency is incorrect, please check it!')
                return self.__class__(self.values[item], self.frequency)
        elif isinstance(item, int):
            return self.values[item]

    def __len__(self):
        return self.length

    # To do: 实现基本的加减乘除功能
    def _check_identity(self, other):
        if self.__class__ != other.__class__:
            return 'Different Domain'
        elif self.frequency != other.frequency:
            return 'Different Frequency'
        elif self.length != other.length:
            return 'Different Length'
        else:
            return 'ok'

    def __add__(self, other):
        if self._check_identity(other) == 'ok':
            return self.__class__(self.values + other.values, self.frequency)
        else:
            raise Exception('Two wavs with ' + self._check_identity(other) + ' can not be added!')
    
    def __sub__(self, other):
        if self._check_identity(other) == 'ok':
            return self.__class__(self.values - other.values, self.frequency)
        else:
            raise Exception('Two wavs with ' + self._check_identity(other) + ' can not be subtracted!')

    def __mul__(self, other):
        if self._check_identity(other) == 'ok':
            return self.__class__(self.values * other.values, self.frequency)
        elif isinstance(other, (int, float)):
            return self.__class__(self.values * other, self.frequency)
        else:
            raise Exception('Two wavs with ' + self._check_identity(other) + ' can not be multiplied!')

    def __div__(self, other):
        pass
    
    # utils
    # 零均值化（去均值化）
    def demean(self, inplace=False):
        if inplace:
            self.values = self.values - np.mean(self.values)
        else:
            return self.__class__(self.values - np.mean(self.values), self.frequency)
    # 归一化
    def normalize(self, inplace=False):
        if inplace:
            self.values = self.values / np.max(np.abs(self.values))
        else:
            return self.__class__(self.values / np.max(np.abs(self.values)), self.frequency)        

    # properties
    # 不依赖任何时间信息或频率信息计算的常用指标
    # 指标全部以属性而不是函数的形式调用
    @cached_property
    def PP(self):
        '''
        最大最小值, 峰峰值PP
        '''
        return np.max(self.values) - np.min(self.values)

    @cached_property
    def Mean(self):
        '''
        平均值(一阶矩)
        '''
        return np.mean(self.values)

    @cached_property
    def RMS(self):
        '''
        均方根值(二阶矩)
        '''
        return np.sqrt(np.mean(np.power(self.values, 2)))

    @cached_property
    def Var(self):
        '''
        方差
        '''
        return np.mean(np.power(self.values - self.Mean, 2))

    @cached_property
    def SD(self):
        '''
        标准差
        '''
        return np.sqrt(self.Var)

    @cached_property
    def Skewness(self):
        '''
        偏度 SK(三阶矩)
        '''
        return np.mean(np.power(self.values - self.Mean, 3)) / np.power(self.SD, 3)
    
    @cached_property
    def IM(self):
        '''
        脉冲指标 IM(Impulse Indicator)
        IM 是信号绝对值的最大值与绝对平均值之商。信号中的冲击越大, IM 越大。
        '''
        return np.max(np.abs(self.values)) / np.mean(np.abs(self.values))

    @cached_property
    def CR(self):
        '''
        峰值指标 CR(Crest Indicator/Factor)
        '''
        return np.max(np.abs(self.values)) / self.RMS

    @cached_property
    def CL(self):
        '''
        裕度指标 CL(Clearance Indicator)
        '''
        return np.max(np.abs(self.values)) / np.power(np.mean(np.sqrt(np.abs(self.values))), 2)

    @cached_property
    def Kurtosis(self):
        '''
        峰度 KU(四阶矩)
        '''
        return np.mean(np.power(self.values - self.Mean, 4)) / np.power(self.SD, 4) - 3

    # 不依赖任何时间和频率信息进行的变换
    def hilbert(self, convert2real=True):
        """
        进行希尔伯特变换

        Parameters  :
        ----------
        convert2real: bool
        是否要将傅里叶变换的结果变为实数(取模)

        Returns  :
        -------
        hilbert_wav: Wav
        经过希尔伯特变换后的波
        """
        if convert2real:
            hilbert = np.abs(signal.hilbert(self.values))
        else:
            hilbert = signal.hilbert(self.values)

        hilbert_wav = self.__class__(hilbert, self.frequency)

        return hilbert_wav

    # 基本绘图函数，后续可对子类定制自己的绘图函数
    def plot(self, *args, **kwargs):
        plt.plot(self.values, *args, **kwargs)
        return None

    # 实现滑动窗计算，每个滑动窗都是一个Wav，可以在每个滑动窗进行所有指标的计算和转换
    def rolling(self, window_width, step=1):
        '''
        Parameters
        ----------
        window_width : int 
        滑动窗的宽度
        step : int
        滑动窗的步长

        Returns
        -------
        rolling_wav: RollingWav
        A generator of Wav
        '''
        rolling_wav = RollingWav(self, window_width, step)
        return rolling_wav
    # 拓展函数，可对数据直接进行自定义的变换和计算
    def apply(self, func, *args, **kwargs):
        self.values = func(self.values, *args, **kwargs)
        return self

if __name__ == "__main__":
    # init basewav
    x = np.linspace(0, 1000, 800000)
    y1 = np.cos(x) 
    y2 = 3 * np.sin(10 * x) + 1
    test_wav1 = BaseWav(y1)
    test_wav2 = BaseWav(y2)
    # test basic function
    print('test basic function'.center(40, '-'))
    print(test_wav1)
    part_wav = test_wav1[100:1000]

    wav_sum = test_wav1 + test_wav2
    wav_diff = test_wav1 - test_wav2
    wav_prod = test_wav1 * test_wav2

    # test utils
    print('test utils'.center(40, '-'))
    test_wav2.demean(inplace=True)

    # test properties
    print('test properties'.center(40, '-'))
    for property in ["PP", "Mean", "RMS", "Var", "SD", "Skewness", "IM", "CR", "CL", "Kurtosis"]:
        print(property +": %.4f" % getattr(wav_sum, property))

    # test tranformation & plot
    print('test transformation'.center(40, '-'))
    wav_prod[:800].plot('r-', label='raw data')
    wav_prod.hilbert()[:800].plot('b-', label='hilbert')
    plt.legend()
    plt.show()

    # test the rolling calculation
    print('test rolling calculation'.center(40, '-'))
    print(wav_sum.rolling(100, 50).RMS)





# -*- encoding: utf-8 -*-
'''
Projects -> File: -> frequencydomainwav.py
Author: DYJ
Date: 2022/07/19 14:44:53
Desc: 频域信号类
      在频域内进行计算的指标放在这个类里
version: 1.0
To do: 与目前系统里的计算方式对比，看结果是否有差异，看实现是否正确
'''

from functools import cached_property

import numpy as np
import matplotlib.pyplot as plt
from scipy import signal

from .basewav import BaseWav
from .rollingwav import RollingWav
from . import timedomainwav as tdwav


class FrequencyDomainWav(BaseWav):
    '''
    频域信号

    在频域内进行计算的指标放在这个类里
    '''
    def __init__(self, values, frequency):
        ''' 
        初始化FDWav, 频域信号初始化必须提供频率信息
        
        Parameters
        ----------
        values : 1D list or array-like
        信号的强度值
        frequency : int
        信号对应的频率

        Returns
        -------
        None
        '''
        self.values = np.asarray(values).squeeze()
        if len(self.values.shape) > 1:
            raise Exception('Wave data can not have more than 1-dimension!')
        self.length = len(self.values)
        self.frequency = frequency

    # utils
    def get_frequency_band(self, lower_bound, higher_bound):
        """
        获取指定频带的数据

        Parameters  :
        ----------
        lower_bound: int or float
        频带下限, 单位Hz
        end_freq: int or float
        频带上限, 单位Hz

        Returns  : 
        -------
        part_wav: FrequencyDomainWav
        指定频带的数据
        
        """
        part_index = (self.frequency >= lower_bound) & (self.frequency < higher_bound)
        part_wav = self.__class__(self.values[part_index], self.frequency[part_index])

        return part_wav

    def find_max_peak(self, start_freq=0, end_freq=np.inf):
        """
        在一定频率范围内查找最大的幅值值及相应的频率
        默认为从0到正无穷查找，这样找出来的是基频
        Parameters  :
        ----------
        start_freq: int or float
        起始频率范围
        end_freq: int or float
        终止频率范围

        Returns  : 
        -------
        (max_peak_frequency, max_peak_magnitude)
        
        """
        
        search_space = self.get_frequency_band(start_freq, end_freq)
        max_peak_frequency = search_space.frequency[np.argmax(search_space.values)]
        max_peak_magnitude = np.max(search_space.values)

        return (max_peak_frequency, max_peak_magnitude)
    # 在频域内计算的指标/依赖频率信息计算的指标
    @cached_property
    def FC(self):
        '''
        重心频率
        '''
        return np.average(self.frequency, weights=self.values)

    @cached_property
    def MSF(self):
        '''
        均方频率
        '''
        return np.average(np.power(self.frequency, 2), weights=self.values)

    @cached_property
    def VF(self):
        '''
        频率方差
        '''
        square_deviate_from_mean = np.power(self.frequency - self.FC, 2)
        return np.average(square_deviate_from_mean, weights=self.values)

    @cached_property
    def RMSF(self):
        '''
        均方根频率
        '''
        return np.sqrt(self.MSF)

    @cached_property
    def RVF(self):
        '''
        频率标准差
        '''
        return np.sqrt(self.VF)
    
    @cached_property
    def baseband(self):
        return self.find_max_peak(5, 95)

    def THD(self, max_harmonic_num=150, leaky_index=1):
        """ 谐波畸变率
        由少青的代码改编而来

        Parameters
        ----------
        max_harmonic_num : int
            最大谐波数
        leaky_index : int
            泄露指数，在谐波附近多大频率范围内寻找谐波峰值
        Returns
        -------
        total_harmonic_distortion : float
            总谐波畸变率

        """
        # 计算基频幅值
        fundamental_frequency = self.frequency[np.argmax(self.values)]
        fundamental_magnitude = np.max(self.values)
        # 计算谐波含有率
        harmonic_component = 0
        # 计算可取的最大谐波数
        available_max_harmonic_num = np.max(self.frequency) // fundamental_frequency - 1
        max_harmonic_num = min(available_max_harmonic_num, max_harmonic_num)
        try :
            for multi in range(2, int(max_harmonic_num) + 1):
                # 指定谐波的搜寻范围
                frequency_series = [multi * fundamental_frequency - leaky_index,
                                    multi * fundamental_frequency + leaky_index]                                
                get_use_fft_data = self.values[(frequency_series[0] < self.frequency) &
                                                (self.frequency < frequency_series[1])]     
                get_use_mag = np.max(get_use_fft_data)
                max_magnitude_square = np.power(get_use_mag, 2)
                harmonic_component += max_magnitude_square
        except ValueError:
            print('The argument max_harmonic_num is too big!')
                
        square_harmonic_content = np.sqrt(harmonic_component)
        total_harmonic_distortion = square_harmonic_content / fundamental_magnitude
        return total_harmonic_distortion

    # 定制绘图函数
    def plot(self, magnitude2log=False, *args, marker=False, marker_num=20, marker_color='r', prominence=2, **kwargs):
        if magnitude2log:
            plt.plot(self.frequency, np.log(self.values), *args, **kwargs)
            plt.xlabel('频率(Hz)', fontdict={'fontsize':16})
            plt.ylabel('幅值(dB)', fontdict={'fontsize':16})
        else:
            plt.plot(self.frequency, self.values, *args, **kwargs)
            plt.xlabel('频率(Hz)', fontdict={'fontsize':16})
            plt.ylabel('幅值', fontdict={'fontsize':16})
        if marker:
            self.mark_extreme_frequency(marker_color, prominence, marker_num, magnitude2log)
        return None
    # 绘图辅助函数
    def mark_extreme_frequency(self, color, prominence, marker_num, magnitude2log):
        '''标记幅值的极值点'''
        if magnitude2log:
            plot_values = np.log(self.values)
            prominence = np.max([prominence, np.std(plot_values)])
        else:
            plot_values = self.values
        peaks_index, peaks_info = signal.find_peaks(plot_values, prominence=prominence)
        if len(peaks_index)==0 :  # 如果没有找到极值点，直接跳出
            print('Peaks not found!')
            return None
        elif len(peaks_index) >= marker_num:  # 如果找到的极值点过多，取前30大极值点进行绘制
            peaks_index = peaks_index[np.argsort(peaks_info['prominences'])[-30:]]
        for idx in peaks_index:
            plt.scatter(self.frequency[idx], plot_values[idx], c=color, s=6)
            plt.text(self.frequency[idx], plot_values[idx], "(%.1f)"%self.frequency[idx], c='k', size=20)
        return None

    # 在频域内进行的计算或变换
    def fft(self, window='hann', convert2real=True):
        '''
        在频域内进行加窗傅里叶变换, 从频域到时域不需要除以数据长度

        Parameters
        ----------
        window : str
            加窗计算所使用的窗的名称, 支持scipy.signal.get_window内的所有窗
        convert2real : bool
            是否将傅里叶变换的结果从复数转化为实数(计算模)
        Returns
        -------
        TDWav
            傅里叶变换后的时域信号
        '''
        # 加窗抑制频谱泄露
        window = signal.get_window(window, self.length)
        windowed_value = self.values * window
        adjust_coefficient = round(1 / np.mean(window), 4)
        if convert2real:
            windowed_value_fft = np.abs(np.fft.fft(windowed_value)) * adjust_coefficient
        else:
            windowed_value_fft = np.fft.fft(windowed_value) * adjust_coefficient

        return tdwav.TimeDomainWav(windowed_value_fft, sample_frequency=1)

# todo: 测试代码
if __name__ == "__main__":
    x = np.linspace(0, 1000, 10000)
    y = np.cos(x) 
    test_wav = FrequencyDomainWav(y, x)
    for property in ["FC", "MSF", "VF", "RMSF", "RVF"]:
        print(property +": %.4f" % getattr(test_wav, property))


import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="wavesis",
    version="0.0.6",
    author="yjdai",
    author_email="136271877@qq.mail.com",
    description="oop data analysis tools for wave data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Blockhead-yj/wavesis.git",
    packages=setuptools.find_packages(),
    classifiers=[],
    python_requires='>=3.7',
)
